/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>

#include "intl.h"
#include "imap.h"
#include "socket.h"
#include "recv.h"
#include "procmsg.h"
#include "procheader.h"
#include "folder.h"
#include "statusbar.h"
#include "prefs_account.h"
#include "codeconv.h"
#include "utils.h"

static GList *session_list = NULL;

static gint imap_cmd_count = 0;

static GSList *imap_get_uncached_messages(IMAPSession	*session,
					  FolderItem	*item,
					  gint		 cache_last,
					  gint		*rfirst,
					  gint		*rlast);
static GSList *imap_delete_old_messages	 (GSList	*mlist,
					  gint		 first);
static void imap_delete_all_messages	 (FolderItem	*item);

static SockInfo *imap_open(const gchar *server, gushort port, gchar *buf);
static gint imap_auth(SockInfo *sock, const gchar *user, const gchar *pass);
static gint imap_logout(SockInfo *sock);
static gint imap_noop(SockInfo *sock);
static gint imap_select(SockInfo *sock, const gchar *folder, gint *num);
static gint imap_get_envelope(SockInfo *sock, gint first, gint last);
static gint imap_get_message(SockInfo *sock, gint num, const gchar *filename);

static gchar *imap_parse_atom(SockInfo *sock, gchar *src, gchar *dest,
			      gchar *orig_buf);
static gchar *imap_parse_one_address(SockInfo *sock, gchar *start,
				     gchar *out_from_str,
				     gchar *out_fromname_str,
				     gchar *orig_buf);
static gchar *imap_parse_address(SockInfo *sock, gchar *start,
				 gchar **out_from_str,
				 gchar **out_fromname_str,
				 gchar *orig_buf);
static MsgInfo *imap_parse_envelope(SockInfo *sock, gchar *line_str);

static gint imap_ok(SockInfo *sock, GPtrArray *argbuf);
static void imap_gen_send(SockInfo *sock, const gchar *format, ...);
static gint imap_gen_recv(SockInfo *sock, gchar *buf, gint size);

static gchar *search_array_contain_str(GPtrArray *array, gchar *str);


Session *imap_session_new(const gchar *server, gushort port,
			  const gchar *user, const gchar *pass)
{
	gchar buf[IMAPBUFSIZE];
	IMAPSession *session;
	SockInfo *imap_sock;

	g_return_val_if_fail(server != NULL, NULL);

	log_message(_("creating IMAP4 connection to %s:%d ...\n"),
		    server, port);

	if ( !(imap_sock = imap_open(server, port, buf)) )
		return NULL;
	if (imap_auth(imap_sock, user, pass) != IMAP_SUCCESS) {
		imap_logout(imap_sock);
		sock_close(imap_sock);
		return NULL;
	}

	session = g_new(IMAPSession, 1);
	SESSION(session)->type      = SESSION_IMAP;
	SESSION(session)->server    = g_strdup(server);
	SESSION(session)->sock      = imap_sock;
	SESSION(session)->connected = TRUE;
	SESSION(session)->phase     = SESSION_READY;
	SESSION(session)->data      = NULL;
	session->mbox = NULL;

	session_list = g_list_append(session_list, session);

	return SESSION(session);
}

void imap_session_destroy(IMAPSession *session)
{
	sock_close(SESSION(session)->sock);

	g_free(session->mbox);

	session_list = g_list_remove(session_list, session);
}

void imap_session_destroy_all(void)
{
	while (session_list != NULL) {
		IMAPSession *session = (IMAPSession *)session_list->data;

		imap_logout(SESSION(session)->sock);
		imap_session_destroy(session);
	}
}

GSList *imap_get_msg_list(Folder *folder, FolderItem *item, gboolean use_cache)
{
	GSList *mlist;
	IMAPSession *session;

	g_return_val_if_fail(folder != NULL, NULL);
	g_return_val_if_fail(item != NULL, NULL);
	g_return_val_if_fail(folder->type == F_IMAP, NULL);
	g_return_val_if_fail(folder->account != NULL, NULL);

	if (!REMOTE_FOLDER(folder)->session) {
		REMOTE_FOLDER(folder)->session =
			imap_session_new(folder->account->recv_server, 143,
					 folder->account->userid,
					 folder->account->passwd);
	} else {
		if (imap_noop(REMOTE_FOLDER(folder)->session->sock)
		    != IMAP_SUCCESS) {
			log_warning(_("IMAP4 connection to %s:%d has been"
				      " disconnected. Reconnecting...\n"),
				    folder->account->recv_server, 143);
			session_destroy(REMOTE_FOLDER(folder)->session);
			REMOTE_FOLDER(folder)->session =
				imap_session_new(folder->account->recv_server,
						 143, folder->account->userid,
						 folder->account->passwd);
		}
	}
	session = IMAP_SESSION(REMOTE_FOLDER(folder)->session);

	if (!session) {
		mlist = procmsg_read_cache(item, FALSE);
		item->last_num = procmsg_get_last_num_in_cache(mlist);
	} else if (use_cache) {
		GSList *newlist;
		gint cache_last;
		gint first, last;

		mlist = procmsg_read_cache(item, FALSE);

		cache_last = procmsg_get_last_num_in_cache(mlist);
		newlist = imap_get_uncached_messages
			(session, item, cache_last, &first, &last);
		mlist = imap_delete_old_messages(mlist, first);

		mlist = g_slist_concat(mlist, newlist);
		item->last_num = last;
	} else {
		gint last;

		mlist = imap_get_uncached_messages
			(session, item, 0, NULL, &last);
		imap_delete_all_messages(item);
		item->last_num = last;
	}

	procmsg_set_flags(mlist, item);

	statusbar_pop_all();

	return mlist;
}

gchar *imap_fetch_msg(Folder *folder, FolderItem *item, gint num)
{
	gchar *path, *filename;
	gint ok;

	g_return_val_if_fail(folder != NULL, NULL);
	g_return_val_if_fail(item != NULL, NULL);

	path = folder_item_get_path(item);
	filename = g_strconcat(path, G_DIR_SEPARATOR_S, itos(num), NULL);
	g_free(path);

	if (is_file_exist(filename)) {
		debug_print(_("message %d has been already cached.\n"), num);
		return filename;
	}

	if (!REMOTE_FOLDER(folder)->session) {
		g_free(filename);
		return NULL;
	}

	debug_print(_("getting message %d...\n"), num);
	ok = imap_get_message(REMOTE_FOLDER(folder)->session->sock,
			      num, filename);
	statusbar_pop_all();
	if (ok != IMAP_SUCCESS) {
		g_warning(_("can't fetch message %d\n"), num);
		g_free(filename);
		return NULL;
	}

	return filename;
}

void imap_scan_folder(Folder *folder, FolderItem *item)
{
}

FolderItem *imap_create_folder(Folder *folder, FolderItem *parent,
			       const gchar *name)
{
	gchar *path;
	FolderItem *new_item;

	g_return_val_if_fail(folder != NULL, NULL);
	g_return_val_if_fail(parent != NULL, NULL);
	g_return_val_if_fail(name != NULL, NULL);

	if (parent->path)
		path = g_strconcat(parent->path, G_DIR_SEPARATOR_S, name,
				   NULL);
	else
		path = g_strdup(name);

	new_item = folder_item_new(name, path);
	folder_item_append(parent, new_item);
	g_free(path);

	return new_item;
}

static GSList *imap_get_uncached_messages(IMAPSession *session,
					  FolderItem *item, gint cache_last,
					  gint *rfirst, gint *rlast)
{
	gint ok;
	gint num = 0, first = 0, last = 0;
	gchar buf[IMAPBUFSIZE];
	GSList *newlist = NULL;
	GSList *llast = NULL;
	MsgInfo *msginfo;

	if (rfirst) *rfirst = 0;
	if (rlast)  *rlast  = 0;

	g_return_val_if_fail(session != NULL, NULL);
	g_return_val_if_fail(item != NULL, NULL);
	g_return_val_if_fail(item->folder != NULL, NULL);
	g_return_val_if_fail(item->folder->type == F_IMAP, NULL);

	ok = imap_select(SESSION(session)->sock, item->path, &num);
	if (ok != IMAP_SUCCESS) {
		log_warning(_("can't set mailbox: %s\n"), item->path);
		return NULL;
	}

	if (num <= cache_last) return NULL;
	if (imap_get_envelope(SESSION(session)->sock, cache_last + 1, num)
	    != IMAP_SUCCESS) {
		log_warning(_("can't get envelope\n"));
		return NULL;
	}

	for (;;) {
		if (sock_read(SESSION(session)->sock, buf, sizeof(buf)) < 0) {
			log_warning(_("error occurred while getting envelope.\n"));
			return newlist;
		}

		if (buf[0] != '*' || buf[1] != ' ') break;

		msginfo = imap_parse_envelope(SESSION(session)->sock, buf);
		if (!msginfo) {
			log_warning(_("can't parse envelope: %s\n"), buf);
			continue;
		}

		msginfo->folder = item;
		msginfo->flags = MSG_NEW|MSG_UNREAD|MSG_IMAP;

		if (!newlist)
			llast = newlist = g_slist_append(newlist, msginfo);
		else {
			llast = g_slist_append(llast, msginfo);
			llast = llast->next;
		}
	}

	if (rfirst) *rfirst = first;
	if (rlast)  *rlast  = last;
	return newlist;
}

static GSList *imap_delete_old_messages(GSList *mlist, gint first)
{
	GSList *cur, *next;
	MsgInfo *msginfo;
	gchar *cache_file;

	if (first < 2) return mlist;

	for (cur = mlist; cur != NULL; ) {
		next = cur->next;

		msginfo = (MsgInfo *)cur->data;
		if (msginfo && msginfo->msgnum < first) {
			debug_print(_("deleting message %d...\n"),
				    msginfo->msgnum);

			cache_file = procmsg_get_message_file_path(msginfo);
			if (is_file_exist(cache_file)) unlink(cache_file);
			g_free(cache_file);

			procmsg_msginfo_free(msginfo);
			mlist = g_slist_remove(mlist, msginfo);
		}

		cur = next;
	}

	return mlist;
}

static void imap_delete_all_messages(FolderItem *item)
{
	DIR *dp;
	struct dirent *d;
	gchar *dir;
	gchar *file;

	g_return_if_fail(item != NULL);
	g_return_if_fail(item->folder != NULL);
	g_return_if_fail(item->folder->type == F_IMAP);

	dir = folder_item_get_path(item);
	if ((dp = opendir(dir)) == NULL) {
		FILE_OP_ERROR(dir, "opendir");
		g_free(dir);
		return;
	}

	debug_print(_("\tDeleting all cached messages... "));

	while ((d = readdir(dp)) != NULL) {
		if (to_number(d->d_name) < 0) continue;

		file = g_strconcat(dir, G_DIR_SEPARATOR_S, d->d_name, NULL);

		if (is_file_exist(file)) {
			if (unlink(file) < 0)
				FILE_OP_ERROR(file, "unlink");
		}

		g_free(file);
	}

	closedir(dp);
	g_free(dir);

	debug_print(_("done.\n"));
}


static SockInfo *imap_open(const gchar *server, gushort port, gchar *buf)
{
	SockInfo *sockinfo;

	if ((sockinfo = sock_connect(server, port)) == NULL) {
		log_warning(_("Can't connect to IMAP4 server: %s:%d\n"),
			    server, port);
		return NULL;
	}

	imap_cmd_count = 0;

	if (imap_noop(sockinfo) != IMAP_SUCCESS) {
		sock_close(sockinfo);
		return NULL;
	}

	return sockinfo;
}

static gint imap_auth(SockInfo *sock, const gchar *user, const gchar *pass)
{
	gint ok;
	GPtrArray *argbuf;

	imap_gen_send(sock, "LOGIN %s %s", user, pass);
	argbuf = g_ptr_array_new();
	ok = imap_ok(sock, argbuf);
	if (ok != IMAP_SUCCESS) log_warning(_("IMAP4 login failed.\n"));
	ptr_array_free_strings(argbuf);
	g_ptr_array_free(argbuf, TRUE);

	return ok;
}

static gint imap_logout(SockInfo *sock)
{
	imap_gen_send(sock, "LOGOUT");
	return imap_ok(sock, NULL);
}

static gint imap_noop(SockInfo *sock)
{
	imap_gen_send(sock, "NOOP");
	return imap_ok(sock, NULL);
}

static gint imap_select(SockInfo *sock, const gchar *folder, gint *num)
{
	gint ok;
	gchar buf[IMAPBUFSIZE];
	gchar *resp_str;
	GPtrArray *argbuf;

	*num = 0;

	imap_gen_send(sock, "SELECT %s", folder);

	argbuf = g_ptr_array_new();
	if ((ok = imap_ok(sock, argbuf)) != IMAP_SUCCESS) {
		ptr_array_free_strings(argbuf);
		g_ptr_array_free(argbuf, TRUE);
		return ok;
	}

	resp_str = search_array_contain_str(argbuf, "EXISTS");
	if (resp_str) {
		if (sscanf(resp_str, "%d %s", num, buf) != 2) {
			g_warning("imap_select(): resp_str is invalid\n");
			ok = IMAP_ERROR;
		} else
			ok = IMAP_SUCCESS;
	}

	ptr_array_free_strings(argbuf);
	g_ptr_array_free(argbuf, TRUE);

	return ok;
}

static gchar *strchr_cpy(const gchar *src, gchar ch, gchar *dest, gint len)
{
	gchar *tmp;

	dest[0] = '\0';
	tmp = strchr(src, ch);
	if (!tmp || tmp == src)
		return NULL;

	memcpy(dest, src, MIN(tmp - src, len - 1));
	dest[MIN(tmp - src, len - 1)] = '\0';

	return tmp + 1;
}

static gint imap_get_message(SockInfo *sock, gint num, const gchar *filename)
{
	gint ok;
	gchar buf[IMAPBUFSIZE];
	gchar *cur_pos;
	gchar size_str[32];
	glong size_num;

	g_return_val_if_fail(filename != NULL, IMAP_ERROR);

	imap_gen_send(sock, "FETCH %d BODY[]", num);

	if (sock_read(sock, buf, sizeof(buf)) < 0)
		return IMAP_ERROR;

	if (buf[0] != '*' || buf[1] != ' ')
		return IMAP_ERROR;
	log_print("IMAP4< %s\n", buf);

	cur_pos = strchr(buf, '{');
	g_return_val_if_fail(cur_pos != NULL, IMAP_ERROR);
	cur_pos = strchr_cpy(cur_pos + 1, '}', size_str, sizeof(size_str));
	g_return_val_if_fail(cur_pos != NULL, IMAP_ERROR);
	size_num = atol(size_str);

	if (strncmp(cur_pos, "\r\n", 2) != 0)
		return IMAP_ERROR;

	if (recv_bytes_write_to_file(sock, size_num, filename) != 0)
		return IMAP_ERROR;

	if (imap_gen_recv(sock, buf, sizeof(buf)) != IMAP_SUCCESS)
		return IMAP_ERROR;

	if (buf[0] != ')')
		return IMAP_ERROR;

	ok = imap_ok(sock, NULL);

	return ok;
}

static gchar *imap_parse_atom(SockInfo *sock, gchar *src, gchar *dest,
			      gchar *orig_buf)
{
	gchar *cur_pos = src;

	while (*cur_pos == ' ') cur_pos++;

	if (!strncmp(cur_pos, "NIL", 3)) {
		*dest = '\0';
		cur_pos += 3;
	} else if (*cur_pos == '\"') {
		gchar *p;

		p = strchr_cpy(cur_pos + 1, '\"', dest, IMAPBUFSIZE);
		cur_pos = p ? p : cur_pos + 2;
	} else if (*cur_pos == '{') {
		gchar buf[32];
		gint len;

		cur_pos = strchr_cpy(cur_pos + 1, '}', buf, sizeof(buf));
		len = atoi(buf);

		g_return_val_if_fail(orig_buf != NULL, cur_pos);

		if (sock_read(sock, orig_buf, IMAPBUFSIZE) < 0)
		log_print("IMAP4< %s\n", orig_buf);
		memcpy(dest, orig_buf, len);
		dest[len] = '\0';
		cur_pos = orig_buf + len;
	}

	return cur_pos;
}

static gchar *imap_parse_one_address(SockInfo *sock, gchar *start,
				     gchar *out_from_str,
				     gchar *out_fromname_str,
				     gchar *orig_buf)
{
	gchar buf[IMAPBUFSIZE];
	gchar *cur_pos = start;

	cur_pos = imap_parse_atom(sock, cur_pos, buf, orig_buf);
	conv_unmime_header(out_fromname_str, 256, buf, NULL);

	if (out_fromname_str[0] != '\0') {
		strcat(out_from_str, "\"");
		strcat(out_from_str, out_fromname_str);
		strcat(out_from_str, "\"");
	}

	cur_pos = imap_parse_atom(sock, cur_pos, buf, orig_buf);

	strcat(out_from_str, " <");

	cur_pos = imap_parse_atom(sock, cur_pos, buf, orig_buf);
	strcat(out_from_str, buf);

	cur_pos = imap_parse_atom(sock, cur_pos, buf, orig_buf);
	strcat(out_from_str, "@");
	strcat(out_from_str, buf);
	strcat(out_from_str, ">");

	while (*cur_pos == ' ') cur_pos++;

	g_return_val_if_fail(*cur_pos == ')', cur_pos + 1);

	return cur_pos + 1;
}

static gchar *imap_parse_address(SockInfo *sock, gchar *start,
				 gchar **out_from_str,
				 gchar **out_fromname_str,
				 gchar *orig_buf)
{
	gchar buf[IMAPBUFSIZE];
	gchar name_buf[IMAPBUFSIZE];
	gchar *cur_pos = start;
	gboolean first = TRUE;

	if (out_from_str)     *out_from_str     = NULL;
	if (out_fromname_str) *out_fromname_str = NULL;
	buf[0] = name_buf[0] = '\0';

	if (!strncmp(cur_pos, "NIL", 3)) {
		if (out_from_str)     *out_from_str     = g_strdup("");
		if (out_fromname_str) *out_fromname_str = g_strdup("");
		return cur_pos + 3;
	}

	g_return_val_if_fail(*cur_pos == '(', NULL);
	cur_pos++;

	for (;;) {
		gchar ch = *cur_pos++;
		if (ch == ')') break;
		if (ch == '(') {
			if (!first) strcat(buf, ", ");
			first = FALSE;
			cur_pos = imap_parse_one_address
				(sock, cur_pos, buf, name_buf, orig_buf);
			if (!cur_pos) return NULL;
		}
	}

	if (out_from_str)     *out_from_str     = g_strdup(buf);
	if (out_fromname_str) *out_fromname_str = g_strdup(name_buf);

	return cur_pos;
}

static MsgInfo *imap_parse_envelope(SockInfo *sock, gchar *line_str)
{
	MsgInfo *msginfo;
	gchar buf[IMAPBUFSIZE];
	gchar tmp[IMAPBUFSIZE];
	gchar *cur_pos;
	gint msgnum;
	size_t size;
	gchar *date = NULL;
	time_t date_t;
	gchar *subject = NULL;
	gchar *tmp_from;
	gchar *tmp_fromname;
	gchar *from = NULL;
	gchar *fromname = NULL;
	gchar *tmp_to;
	gchar *to = NULL;
	gchar *inreplyto = NULL;
	gchar *msgid = NULL;

	log_print("IMAP4< %s\n", line_str);

	g_return_val_if_fail(line_str != NULL, NULL);
	g_return_val_if_fail(line_str[0] == '*' && line_str[1] == ' ', NULL);

	cur_pos = line_str + 2;

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ' ', buf, sizeof(buf))), NULL);
	msgnum = atoi(buf);

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ' ', buf, sizeof(buf))), NULL);
	g_return_val_if_fail(!strcmp(buf, "FETCH"), NULL);

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ' ', buf, sizeof(buf))), NULL);
	g_return_val_if_fail(!strcmp(buf, "(FLAGS"), NULL);

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ')', buf, sizeof(buf))), NULL);

	g_return_val_if_fail(*cur_pos == ' ', NULL);
	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos + 1, ' ', buf, sizeof(buf))),
		 NULL);
	g_return_val_if_fail(!strcmp(buf, "RFC822.SIZE"), NULL);

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ' ', buf, sizeof(buf))), NULL);
	size = atoi(buf);

	g_return_val_if_fail
		((cur_pos = strchr_cpy(cur_pos, ' ', buf, sizeof(buf))), NULL);
	g_return_val_if_fail(!strcmp(buf, "ENVELOPE"), NULL);

	g_return_val_if_fail(*cur_pos == '(', NULL);
	cur_pos = imap_parse_atom(sock, cur_pos + 1, buf, line_str);
	Xstrdup_a(date, buf, return NULL);
	date_t = procheader_date_parse(NULL, date, 0);

	cur_pos = imap_parse_atom(sock, cur_pos, buf, line_str);
	if (buf[0] != '\0') {
		conv_unmime_header(tmp, sizeof(tmp), buf, NULL);
		Xstrdup_a(subject, tmp, return NULL);
	}

	g_return_val_if_fail(*cur_pos == ' ', NULL);
	cur_pos = imap_parse_address(sock, cur_pos + 1,
				     &tmp_from, &tmp_fromname, line_str);
	Xstrdup_a(from, tmp_from,
		  {g_free(tmp_from); g_free(tmp_fromname); return NULL;});
	Xstrdup_a(fromname, tmp_fromname,
		  {g_free(tmp_from); g_free(tmp_fromname); return NULL;});
	g_free(tmp_from);
	g_free(tmp_fromname);

#define SKIP_ONE_ELEMENT() \
{ \
	g_return_val_if_fail(*cur_pos == ' ', NULL); \
	cur_pos = imap_parse_address(sock, cur_pos + 1, \
				     NULL, NULL, line_str); \
}

	/* skip sender and reply-to */
	SKIP_ONE_ELEMENT();
	SKIP_ONE_ELEMENT();

	g_return_val_if_fail(*cur_pos == ' ', NULL);
	cur_pos = imap_parse_address(sock, cur_pos + 1, &tmp_to, NULL, line_str);
	Xstrdup_a(to, tmp_to, {g_free(tmp_to); return NULL;});
	g_free(tmp_to);

	/* skip Cc and Bcc */
	SKIP_ONE_ELEMENT();
	SKIP_ONE_ELEMENT();

	g_return_val_if_fail(*cur_pos == ' ', NULL);
	cur_pos = imap_parse_atom(sock, cur_pos, buf, line_str);
	if (buf[0] != '\0') {
		eliminate_parenthesis(buf, '(', ')');
		extract_parenthesis(buf, '<', '>');
		remove_space(buf);
		Xstrdup_a(inreplyto, buf, return NULL);
	}

	g_return_val_if_fail(*cur_pos == ' ', NULL);
	cur_pos = imap_parse_atom(sock, cur_pos, buf, line_str);
	if (buf[0] != '\0') {
		extract_parenthesis(buf, '<', '>');
		remove_space(buf);
		Xstrdup_a(msgid, buf, return NULL);
	}

	msginfo = g_new0(MsgInfo, 1);
	msginfo->msgnum = msgnum;
	msginfo->size = size;
	msginfo->date = g_strdup(date);
	msginfo->date_t = date_t;
	msginfo->subject = g_strdup(subject);
	msginfo->from = g_strdup(from);
	msginfo->fromname = g_strdup(fromname);
	msginfo->to = g_strdup(to);
	msginfo->inreplyto = g_strdup(inreplyto);
	msginfo->msgid = g_strdup(msgid);

	return msginfo;
}

gint imap_get_envelope(SockInfo *sock, gint first, gint last)
{
	imap_gen_send(sock, "FETCH %d:%d (FLAGS RFC822.SIZE ENVELOPE)",
		      first, last);

	return IMAP_SUCCESS;
}

static gint imap_ok(SockInfo *sock, GPtrArray *argbuf)
{
	gint ok;
	gchar buf[IMAPBUFSIZE];
	gint cmd_num;
	gchar cmd_status[IMAPBUFSIZE];

	while ((ok = imap_gen_recv(sock, buf, sizeof(buf))) == IMAP_SUCCESS) {
		if (buf[0] == '*' && buf[1] == ' ') {
			if (argbuf)
				g_ptr_array_add(argbuf, g_strdup(&buf[2]));
		} else {
			if (sscanf(buf, "%d %s", &cmd_num, cmd_status) < 2)
				return IMAP_ERROR;
			else if (cmd_num == imap_cmd_count &&
				 !strcmp(cmd_status, "OK")) {
				if (argbuf)
					g_ptr_array_add(argbuf, g_strdup(buf));
				return IMAP_SUCCESS;
			} else
				return IMAP_ERROR;
		}
	}

	return ok;
}

static gchar *search_array_contain_str(GPtrArray *array, gchar *str)
{
	gint i;

	for (i = 0; i < array->len; i++) {
		gchar *tmp = g_ptr_array_index(array, i);
		if (strstr(tmp, str) != NULL)
			return tmp;
	}

	return NULL;
}

static void imap_gen_send(SockInfo *sock, const gchar *format, ...)
{
	gchar buf[IMAPBUFSIZE];
	gchar tmp[IMAPBUFSIZE];
	gchar *p;
	va_list args;

	va_start(args, format);
	g_vsnprintf(tmp, sizeof(tmp), format, args);
	va_end(args);

	imap_cmd_count++;

	g_snprintf(buf, sizeof(buf), "%d %s\r\n", imap_cmd_count, tmp);
	if (!strncasecmp(tmp, "LOGIN ", 6) && (p = strchr(tmp + 6, ' '))) {
		*(p + 1) = '\0';
		log_print("IMAP4> %d %s ********\n", imap_cmd_count, tmp);
	} else
		log_print("IMAP4> %d %s\n", imap_cmd_count, tmp);

	sock_write(sock, buf, strlen(buf));
}

static gint imap_gen_recv(SockInfo *sock, gchar *buf, gint size)
{
	if (sock_read(sock, buf, size) == -1)
		return IMAP_SOCKET;

	strretchomp(buf);

	log_print("IMAP4< %s\n", buf);

	return IMAP_SUCCESS;
}
