/*    Windows version of sockets.c using channels.
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef HAVE_DOSISH_SYSTEM

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <windows.h>

#include "socket.h"

#define BUFFSIZE	8192

static const char*
sock_strerror ( int n )
{
    static char buf[15];

    if (n == -1 )
        n = WSAGetLastError ();

    switch(n) {
      case WSAEINTR           : return "intr";
      case WSAEBADF           : return "badf";
      case WSAEACCES          : return "acces";
      case WSAEFAULT          : return "fault";
      case WSAEINVAL          : return "inval";
      case WSAEMFILE          : return "mfile";
      case WSAEWOULDBLOCK     : return "wouldblock";
      case WSAEINPROGRESS     : return "inprogress";
      case WSAEALREADY        : return "already";
      case WSAENOTSOCK        : return "not a socket";
      case WSAEDESTADDRREQ    : return "destaddrreq";
      case WSAEMSGSIZE        : return "msgsize";
      case WSAEPROTOTYPE      : return "prototype";
      case WSAENOPROTOOPT     : return "noprotoopt";
      case WSAEPROTONOSUPPORT : return "protonosupport";
      case WSAESOCKTNOSUPPORT : return "socktnosupport";
      case WSAEOPNOTSUPP      : return "opnotsupp";
      case WSAEPFNOSUPPORT    : return "pfnosupport";
      case WSAEAFNOSUPPORT    : return "afnosupport";
      case WSAEADDRINUSE      : return "addrinuse";
      case WSAEADDRNOTAVAIL   : return "addrnotavail";
      case WSAENETDOWN        : return "net down";
      case WSAENETUNREACH     : return "net unreachable";
      case WSAENETRESET       : return "net reset";
      case WSAECONNABORTED    : return "conn aborted";
      case WSAECONNRESET      : return "conn reset";
      case WSAENOBUFS         : return "nobufs";
      case WSAEISCONN         : return "is connected";
      case WSAENOTCONN        : return "not connectd";
      case WSAESHUTDOWN       : return "shutdown";
      case WSAETOOMANYREFS    : return "toomanyrefs";
      case WSAETIMEDOUT       : return "timed out";
      case WSAECONNREFUSED    : return "connection refused";
      case WSAELOOP           : return "loop";
      case WSAENAMETOOLONG    : return "name too long";
      case WSAEHOSTDOWN       : return "host down";
      case WSAEHOSTUNREACH    : return "host unreachable";
      case WSAENOTEMPTY       : return "not empty";
      case WSAEPROCLIM        : return "proclim";
      case WSAEUSERS          : return "users";
      case WSAEDQUOT          : return "dquot";
      case WSAESTALE          : return "stale";
      case WSAEREMOTE         : return "remote";
      case WSASYSNOTREADY     : return "sys not ready";
      case WSAVERNOTSUPPORTED : return "ver not supported";
      case WSANOTINITIALISED  : return "not initialised";
      case WSAHOST_NOT_FOUND  : return "host not found";
      case WSATRY_AGAIN       : return "host not found (non-authoritative)";
      case WSANO_RECOVERY     : return "not recovery";
      case WSANO_DATA         : return "no data";
      default: sprintf(buf, "rc=%d", n ); return buf;
    }
}    


static void
init_sockets (void)
{
    static int initialized;
    static WSADATA wsdata;

    if (initialized)
        return;

    if( WSAStartup( 0x0101, &wsdata ) ) {
        fprintf ( stdout, "error initializing socket library: %s\n", 
                  sock_strerror (-1) );
        return;
    }
    if( wsdata.wVersion < 0x0001 ) {
        fprintf (stdout, "socket library version is %x.%x - but 1.1 needed\n",
                 LOBYTE(wsdata.wVersion), HIBYTE(wsdata.wVersion));
        WSACleanup();
        return;
    }
    initialized = 1;
}

static gint 
sock_connect_by_hostname(gint sock, const gchar *hostname, gushort port)
{
    struct hostent *hp;
    struct sockaddr_in ad;
    unsigned long l;
    
    memset (&ad, 0, sizeof(ad));
    ad.sin_family = AF_INET;
    ad.sin_port = htons(port);

    if( (l = inet_addr (hostname)) != SOCKET_ERROR ) {
        memcpy (&ad.sin_addr, &l, sizeof(l));
    }
    else if( (hp = gethostbyname (hostname)) ) {
        if( hp->h_addrtype != AF_INET ) {
            fprintf (stdout, "%s: unknown address family\n", hostname);
            return -1;
        }
        if ( hp->h_length != 4 ) {
            fprintf ( stdout, "%s: illegal address length\n", hostname);
            return -1;
        }
        memcpy (&ad.sin_addr, hp->h_addr, hp->h_length);
    }
    else {
        fprintf(stdout, "%s: %s\n", hostname, sock_strerror (-1));
        return -1;
    }

    if( connect (sock, (struct sockaddr *)&ad, sizeof (ad) ) ) {
        fprintf(stdout, "%s: connect problem: %s\n",
                hostname, sock_strerror (-1));
        return -1;
    }

    return 0;
}


static SockInfo *
do_connect (const gchar *hostname, gushort port, gboolean nonblock)
{
    gint sock;
    gint ret;
    SockInfo *sockinfo;
    int inprog;
    GIOChannel *channel;

    fprintf ( stdout, "** sock_connect%s\n", nonblock?"_nb":""); 
    init_sockets ();
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
        fprintf(stdout, "error creating socket: %s\n", sock_strerror (-1));
        return NULL;
    }
    if ( !(channel = g_io_channel_win32_new_stream_socket (sock)) ) {
        fprintf (stdout, "failed to create a channel for a socket\n");
        closesocket (sock);
        return NULL;
    }

    if (nonblock) {
        /* Hmmm: Do we really need this? */
       /*    if (sock_set_nonblocking_mode(sock, TRUE) < 0) 
             return NULL; */
    }

    inprog = 0;
    ret = sock_connect_by_hostname(sock, hostname, port);
    if (ret < 0 ) {
        inprog = (WSAGetLastError () == WSAEWOULDBLOCK);
        if (!inprog) {
            g_io_channel_unref (channel);
            closesocket (sock);
            return NULL;
        }
    }

    sockinfo = g_new0(SockInfo, 1);
    sockinfo->channel = channel;
    sockinfo->hostname = g_strdup(hostname);
    sockinfo->port = port;
    sockinfo->state = inprog? CONN_LOOKUPSUCCESS : CONN_ESTABLISHED;

    return sockinfo;
}

SockInfo *
sock_connect_nb (const gchar *hostname, gushort port)
{
    return do_connect (hostname, port, TRUE);
}

SockInfo *
sock_connect (const gchar *hostname, gushort port)
{
    return do_connect (hostname, port, FALSE);
}


gint 
sock_close (SockInfo *sockinfo)
{
    if (!sockinfo)
        return 0;
    g_io_channel_close (sockinfo->channel);
    g_io_channel_unref (sockinfo->channel);
    sockinfo->channel = NULL;
    g_free(sockinfo->hostname);
    g_free(sockinfo);
    fflush (stdout);
    return 0;
}


gint 
sock_write(SockInfo *sockinfo, const gchar *buf, gint len)
{
    int nwritten;
    GIOError err;
    guint bytes_written;

    g_return_val_if_fail (sockinfo, -1);
    g_return_val_if_fail (sockinfo->channel, -1);

    if (!len)
        return 0;

    err = g_io_channel_write (sockinfo->channel, buf, len, &bytes_written);
    if (err) {
        fprintf (stdout, "channel write problem: %d\n", err);
        return -1;
    }
    nwritten = bytes_written;

    if( !nwritten ) {
        fprintf (stdout, "eof while writing to socket\n");
        return -1;
    }
    if (nwritten != len) {
        fprintf (stdout, "wrote only %d of %d bytes to socket\n",
                nwritten, len);
        return -1;
    }

    return nwritten;
}


static void
push_bytes (SockInfo *sockinfo, const char *buf, int len )
{
    int i;
    char *p = g_malloc ( sockinfo->buflen + len );

    if( sockinfo->buflen ) 
        memcpy (p, sockinfo->buf, sockinfo->buflen );
    memcpy (p+sockinfo->buflen, buf, len );
    sockinfo->buflen += len;
    if (sockinfo->buf)
        g_free (sockinfo->buf);
    sockinfo->buf = p;
}


/* we try to return only one line */
static int
pop_bytes (SockInfo *sockinfo, char *buf, int maxlen )
{
    int i;
    char *nl;
    
    if (!sockinfo->buflen) {
        return 0;
    }
        
    if (sockinfo->buflen < maxlen)
        maxlen = sockinfo->buflen;
    nl = memchr(sockinfo->buf, '\n', maxlen);
    if (nl)
        maxlen = nl - sockinfo->buf + 1;
    
    memcpy (buf, sockinfo->buf, maxlen );
    memmove (sockinfo->buf, sockinfo->buf+maxlen, sockinfo->buflen-maxlen );
    sockinfo->buflen -= maxlen;
    return maxlen;
}


gboolean
sock_bytes_pending (SockInfo *sockinfo)
{
    return sockinfo && sockinfo->buflen;
}

gboolean
sock_lines_pending (SockInfo *sockinfo)
{
    return sockinfo && sockinfo->buflen 
           && memchr(sockinfo->buf, '\n', sockinfo->buflen);
}


gint 
sock_read (SockInfo *sockinfo, gchar *buf, gint len)
{
    gchar *newline, *bp = buf;
    gint n;
    GIOError err;
    guint bytes_read, rest;

    g_return_val_if_fail (sockinfo, -1);
    g_return_val_if_fail (sockinfo->channel, -1);

    if (--len < 1) {
        fprintf (stdout, "** sock read with 0 bytes buffer\n");
        return -1;
    }

    /* read until we found a newline or the buffer gets too short */
    do {
        bytes_read = pop_bytes (sockinfo, bp, len );
        if( !bytes_read ) {
            err = g_io_channel_read (sockinfo->channel,
                                     bp, len, &bytes_read);
            if (err) {
                fprintf (stdout, "channel read problem: %d\n", err);
                return -1;
            }
        }
        n = bytes_read;
        if ( !bytes_read ) {
            fprintf (stdout, "eof encountered at socket\n");
            return -1;
        }
        if ( (newline = memchr(bp, '\n', n)) ) {
            rest = n;
            n = newline - bp + 1;
            rest -= n;
        }
        bp += n;
        len -= n;
    } while (!newline && len);
    
    if (newline) {
        /* save the rest of the buffer */
        push_bytes (sockinfo, bp, rest);
    }

    *bp = '\0';
    return bp - buf;
}



gint 
sock_printf(SockInfo *sockinfo, const gchar *format, ...)
{
    va_list args;
    gchar buf[BUFFSIZE];
    
    va_start(args, format);
    g_vsnprintf(buf, sizeof(buf), format, args);
    va_end(args);
    
    return sock_write(sockinfo, buf, strlen(buf));
}


gint 
sock_puts(SockInfo *sockinfo, const gchar *buf)
{
    gint ret;
    
    if ((ret = sock_write(sockinfo, buf, strlen(buf))) < 0)
        return ret;
    return sock_write(sockinfo, "\r\n", 2);
}

#endif /*HAVE_DOSISH_SYSTEM*/




